unit MLParse;
// Custom XML-like Markup Language Parser (No, this is *NOT* a real XML parser!)
// (C) 2003 Martin J. Fiedler <martin.fiedler@gmx.net>
// This is free software. You may freely redistribute and modify it.
interface

// TMLAttribute: record of a single attribute
type PMLAttribute=^TMLAttribute;
     TMLAttribute=record
                    Name,Value:string;
                  end;
const EmptyMLAttribute:TMLAttribute=(Name:'';Value:'');

// TMLTag: class that represents a single ML tag with its attributes and subtags
type PMLTag=^TMLTag;
     TMLTag=class
            private
              FName:string;  // tag name
              FAttributes:array of TMLAttribute;  // attributes
              FSubItems:array of TMLTag;  // subtags
              // property interfaces
              function GetAttribute(Index:integer):TMLAttribute;
              function GetAttributeCount:integer;
              function GetSubItem(Index:integer):TMLTag;
              function GetSubItemCount:integer;
            public
              // initialization, finalization, basic properties
              constructor Create(const Name:string);
              destructor Destroy; override;
              property Name:string read FName;
              // dumping (=saving) and parsing (=loading)
              function DumpToString(Indent:integer):string;
              function DumpToFile(const FileName:string):boolean;
              procedure ParseFromString(const s:string);
              function ParseFromFile(const FileName:string):boolean;
              // Attributes
              property Attributes[Index:integer]:TMLAttribute read GetAttribute;
              property AttributeCount:integer read GetAttributeCount;
              function GetAttributeByName(const Name:string):string;
              procedure AddAttribute(const _Name,_Value:string);
              // SubItems
              property SubItems[Index:integer]:TMLTag read GetSubItem;
              property SubItemCount:integer read GetSubItemCount;
              function GetSubItemByName(const Name:string):TMLTag;
              procedure AddSubItem(Tag:TMLTag);
              function CreateSubItem(const Name:string):TMLTag;
            end;

// tool routines for encoding attribute values in a HTML-like manner (i.e.
// "=&quot; &=&amp; <=%lt; >=&gt;)
function MLCharEncode(const MyText:string):string;
function MLCharDecode(const Encoded:string):string;
// normally, the user does not need to call these directly, because the Dump*
// and Parse* encode or decode attribute values automatically

implementation
uses SysUtils;

function MLCharEncode(const MyText:string):string;
var Index:integer;
  procedure repl(const by:string);
  begin
    Delete(Result,Index,1);
    Insert(by,Result,Index);
    inc(Index,length(by));
  end;
begin
  Result:=MyText; Index:=1;
  while Index<=length(Result) do
    case Result[Index] of
      '&':repl('&amp;');
      '"':repl('&quot;');
      '<':repl('&lt;');
      '>':repl('&gt;');
    else
      inc(Index);
    end;
end;

function MLCharDecode(const Encoded:string):string;
begin
  Result:=StringReplace(StringReplace(StringReplace(StringReplace(Encoded,
          '&gt;','>',[rfReplaceAll]),'&lt;','>',[rfReplaceAll]),
          '&quot;',#34,[rfReplaceAll]),'&amp;','&',[rfReplaceAll]);
end;

constructor TMLTag.Create(const Name:string);
begin
  inherited Create;
  FName:=Name;
  SetLength(FAttributes,0);
  SetLength(FSubItems,0);
end;

destructor TMLTag.Destroy;
var Index:integer;
begin
  Finalize(FAttributes);
  for Index:=0 to High(FSubItems) do
    FreeAndNil(FSubItems[Index]);
  Finalize(FSubItems);
  inherited Destroy;
end;

function TMLTag.DumpToString(Indent:integer):string;
var Index:integer;
begin
  Result:=StringOfChar(#32,Indent)+'<'+FName;
  for Index:=0 to High(FAttributes) do
    with FAttributes[Index] do
      Result:=Result+#32+Name+'='#34+MLCharEncode(Value)+#34;
  if length(FSubItems)=0 then
    Result:=Result+' />'^M^J
  else begin
    Result:=Result+'>'^M^J;
    for Index:=0 to High(FSubItems) do
      Result:=Result+FSubItems[Index].DumpToString(Indent+1);
    Result:=Result+StringOfChar(#32,Indent)+'</'+FName+'>'^M^J;
  end;
end;

function TMLTag.GetAttribute(Index:integer):TMLAttribute;
begin
  if (Index>=0) AND (Index<length(FAttributes))
    then Result:=FAttributes[Index]
    else Result:=EmptyMLAttribute;
end;

function TMLTag.GetAttributeCount:integer;
begin
  Result:=length(FAttributes);
end;

function TMLTag.GetAttributeByName(const Name:string):string;
var Index:integer;
begin
  for Index:=0 to High(FAttributes) do
    if FAttributes[Index].Name=Name then begin
      Result:=FAttributes[Index].Value;
      exit;
    end;
  Result:='';
end;

procedure TMLTag.AddAttribute(const _Name,_Value:string);
begin
  SetLength(FAttributes,length(FAttributes)+1);
  with FAttributes[High(FAttributes)] do begin
    Name:=_Name;
    Value:=_Value;
  end;
end;

function TMLTag.GetSubItem(Index:integer):TMLTag;
begin
  if (Index>=0) AND (Index<length(FSubItems))
    then Result:=FSubItems[Index]
    else Result:=nil;
end;

function TMLTag.GetSubItemCount:integer;
begin
  Result:=length(FSubItems);
end;

function TMLTag.GetSubItemByName(const Name:string):TMLTag;
var Index:integer;
begin
  for Index:=0 to High(FSubItems) do
    if FSubItems[Index].FName=Name then begin
      Result:=FSubItems[Index];
      exit;
    end;
  Result:=nil;
end;

procedure TMLTag.AddSubItem(Tag:TMLTag);
begin
  if not Assigned(Tag) then exit;
  SetLength(FSubItems,length(FSubItems)+1);
  FSubItems[High(FSubItems)]:=Tag;
end;

function TMLTag.CreateSubItem(const Name:string):TMLTag;
begin
  Result:=TMLTag.Create(Name);
  AddSubItem(Result);
end;

////////////////////////////////////////////////////////////////////////////////
function NextSeparator(const s:string):integer;
var i:integer; InsideQuote:boolean;
begin
  InsideQuote:=false;
  for i:=1 to length(s) do
    case s[i] of
      #0..#32:if not InsideQuote then begin
                Result:=i; exit;
              end;
      #34:InsideQuote:=not(InsideQuote);
    end;
  Result:=length(s)+1;
end;

function ParseSingleTag(var s:string; var Continue:boolean):TMLTag;
var Start,Stop:integer;
    Tag,TagName,Attrib:string;
    NextSep,Equals:integer;
    TagEndsHere:boolean;
begin
  Result:=nil; Continue:=false;
  // find first starting tag
  Start:=pos('<',s); Stop:=pos('>',s);
  if (Start=0) OR (Stop<=Start) then exit;
  // copy the starting tag
  Tag:=Trim(copy(s,Start+1,Stop-Start-1));
  if (length(Tag)<1) OR (Tag[1] IN ['!','?','/']) then begin
    Delete(s,1,Stop); Continue:=true; exit;
  end;
  // is it an end tag?
  TagEndsHere:=(Tag[length(Tag)]='/');
  if TagEndsHere then Tag:=TrimRight(copy(Tag,1,length(Tag)-1));
  // get the tag name
  NextSep:=NextSeparator(Tag);
  TagName:=copy(Tag,1,NextSep-1);
  Result:=TMLTag.Create(TagName);
  Tag:=TrimLeft(copy(Tag,NextSep+1,length(Tag)));
  // fill in attributes
  while Tag<>'' do begin
    NextSep:=NextSeparator(Tag);
    Attrib:=copy(Tag,1,NextSep-1);
    // find '=' in attribute
    Equals:=pos('=',Attrib);
    if Equals=0 then
      Result.AddAttribute(Attrib,Attrib)  // valueless true attribute
    else if Equals<length(Attrib) then begin
      // add attribute and trim quotes if necessary
      if (Attrib[Equals+1]=#34) AND (Attrib[length(Attrib)]=#34)
        then Result.AddAttribute(copy(Attrib,1,Equals-1),
                    MLCharDecode(copy(Attrib,Equals+2,length(Attrib)-Equals-2)))
        else Result.AddAttribute(copy(Attrib,1,Equals-1),
                    MLCharDecode(copy(Attrib,Equals+1,length(Attrib)-Equals)));
    end;
    Tag:=TrimLeft(copy(Tag,NextSep+1,length(Tag)));
  end;
  // stop here if it is an end tag
  if TagEndsHere then begin
    Delete(s,1,Stop); Continue:=true; exit;
  end;
  // otherwise, search for the end tag elsewhere and parse the inner region
  NextSep:=pos('</'+TagName+'>',s);
  if NextSep=0 then begin
    Delete(s,1,Stop); Continue:=true; exit;
  end;
  Result.ParseFromString(copy(s,Stop+1,NextSep-Stop-1));
  Delete(s,1,NextSep+length(TagName)+2); Continue:=true;
end;

procedure TMLTag.ParseFromString(const s:string);
var MyString:string;
    Cont:boolean;
begin
  MyString:=s; Cont:=true;
  while Cont do
    AddSubItem(ParseSingleTag(MyString,Cont));
end;

function TMLTag.ParseFromFile(const FileName:string):boolean;
var F:file; S:string;
    OldFileMode:byte;
    Size:integer;
begin
  try
    AssignFile(F,FileName);
    OldFileMode:=FileMode; FileMode:=0;
    Reset(F,1);
    FileMode:=OldFileMode;
    Size:=FileSize(F);
    SetLength(S,Size);
    BlockRead(F,S[1],Size);
    CloseFile(F);
    ParseFromString(S);
    Result:=true;
  except
    Result:=false;
  end;
end;

function TMLTag.DumpToFile(const FileName:string):boolean;
var F:file; S:string;
begin
  try
    S:=DumpToString(0);
    AssignFile(F,FileName);
    Rewrite(F,1);
    BlockWrite(F,S[1],length(S));
    CloseFile(F);
    Result:=true;
  except
    Result:=false;
  end;
end;

end.
